from enum import Enum
from typing import Dict, Optional, Type

from pydantic import BaseModel


class AutoValueEnum(Enum):
    def _generate_next_value_(name, start, count, last_values):
        return name



from typing import List

from pydantic import BaseModel

from ..collection import Collection
from ..links import Link


class Collections(BaseModel):
    """
    http://docs.opengeospatial.org/is/17-069r3/17-069r3.html#_feature_collections_rootcollections
    """

    links: List[Link]
    collections: List[Collection]

    def to_dict(self, **kwargs):
        return self.dict(by_alias=True, exclude_unset=True, **kwargs)

    def to_json(self, **kwargs):
        return self.json(by_alias=True, exclude_unset=True, **kwargs)



from typing import List, Optional, Union

from pydantic import AnyUrl, BaseModel, Field, constr

from stac_pydantic.links import Links
from stac_pydantic.version import STAC_VERSION


class LandingPage(BaseModel):
    """
    https://github.com/radiantearth/stac-api-spec/blob/master/api-spec.md#ogc-api---features-endpoints
    """

    id: constr(min_length=1)
    description: constr(min_length=1)
    title: Optional[str]
    stac_version: str = Field(STAC_VERSION, const=True)
    stac_extensions: Optional[List[AnyUrl]]
    conformsTo: List[AnyUrl]
    links: Links
    type: constr(min_length=1) = Field("Catalog", const=True)



from enum import auto

from pydantic import BaseModel, constr

from stac_pydantic.utils import AutoValueEnum


class SortDirections(str, AutoValueEnum):
    asc = auto()
    desc = auto()


class SortExtension(BaseModel):
    """
    https://github.com/radiantearth/stac-api-spec/tree/master/extensions/sort#sort-api-extension
    """

    field: constr(min_length=1)
    direction: SortDirections



from enum import auto
from typing import Any, Dict, Iterator, List, Optional, Union
from urllib.parse import urljoin

from pydantic import BaseModel, Field, constr

from stac_pydantic.utils import AutoValueEnum


class PaginationMethods(str, AutoValueEnum):
    """
    https://github.com/radiantearth/stac-api-spec/blob/master/api-spec.md#paging-extension
    """

    GET = auto()
    POST = auto()


class PaginationRelations(str, AutoValueEnum):
    """
    https://github.com/radiantearth/stac-api-spec/blob/master/api-spec.md#paging-extension
    """

    next = auto()
    previous = auto()


class Link(BaseModel):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/collection-spec/collection-spec.md#link-object
    """

    href: constr(min_length=1)
    rel: constr(min_length=1)
    type: Optional[str]
    title: Optional[str]
    # Label extension
    label: Optional[str] = Field(None, alias="label:assets")

    class Config:
        use_enum_values = True

    def resolve(self, base_url: str):
        """resolve a link to the given base URL"""
        self.href = urljoin(base_url, self.href)


class PaginationLink(Link):
    """
    https://github.com/radiantearth/stac-api-spec/blob/master/api-spec.md#paging-extension
    """

    rel: PaginationRelations
    method: PaginationMethods
    body: Optional[Dict[Any, Any]]
    merge: bool = False


class Links(BaseModel):
    __root__: List[Union[PaginationLink, Link]]

    def resolve(self, base_url: str):
        """resolve all links to the given base URL"""
        for link in self:
            link.resolve(base_url)

    def append(self, link: Link):
        self.__root__.append(link)

    def __iter__(self) -> Iterator[Link]:
        """iterate through links"""
        return iter(self.__root__)

    def __len__(self):
        return len(self.__root__)

    def __getitem__(self, idx):
        return self.__root__[idx]


class Relations(str, AutoValueEnum):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/collection-spec/collection-spec.md#relation-types
    """

    self = auto()
    root = auto()
    parent = auto()
    child = auto()
    item = auto()
    license = auto()
    derived_from = auto()
    collection = auto()
    alternate = auto()
    previous = auto()
    next = auto()
    conformance = auto()
    docs = auto()
    tiles = auto()
    search = auto()
    preview = auto()



from functools import singledispatch

import jsonschema
import requests

from stac_pydantic import Catalog, Collection, Item


@singledispatch
def validate_extensions(stac_obj, b):
    raise NotImplementedError("Unsupported type")


@validate_extensions.register
def _(stac_obj: dict, reraise_exception: bool = False, **kwargs) -> bool:
    try:
        if stac_obj["stac_extensions"]:
            for ext in stac_obj["stac_extensions"]:
                req = requests.get(ext)
                schema = req.json()
                jsonschema.validate(instance=stac_obj, schema=schema)
    except Exception:
        if reraise_exception:
            raise
        return False
    return True


@validate_extensions.register
def _(stac_obj: Item, reraise_exception: bool = False, **kwargs) -> bool:
    validate_extensions(stac_obj.dict(), reraise_exception)


@validate_extensions.register
def _(stac_obj: Collection, reraise_exception: bool = False, **kwargs) -> bool:
    validate_extensions(stac_obj.dict(), reraise_exception)


@validate_extensions.register
def _(stac_obj: Catalog, reraise_exception: bool = False, **kwargs) -> bool:
    validate_extensions(stac_obj.dict(), reraise_exception)



from datetime import datetime
from typing import Any, Dict, List, Optional, Union

from geojson_pydantic.geometries import (
    LineString,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
    _GeometryBase,
)
from pydantic import BaseModel, Field, validator
from pydantic.datetime_parse import parse_datetime

from stac_pydantic.api.extensions.fields import FieldsExtension
from stac_pydantic.api.extensions.query import Operator
from stac_pydantic.api.extensions.sort import SortExtension
from stac_pydantic.shared import BBox


class Search(BaseModel):
    """
    https://github.com/radiantearth/stac-api-spec/blob/master/api-spec.md#filter-parameters-and-fields
    """

    collections: List[str]
    ids: Optional[List[str]]
    bbox: Optional[BBox]
    intersects: Optional[
        Union[Point, MultiPoint, LineString, MultiLineString, Polygon, MultiPolygon]
    ]
    datetime: Optional[str]
    limit: int = 10
    field: Optional[FieldsExtension] = Field(None, alias="fields")
    query: Optional[Dict[str, Dict[Operator, Any]]]
    sortby: Optional[List[SortExtension]]

    @property
    def start_date(self) -> Optional[datetime]:
        values = self.datetime.split("/")
        if len(values) == 1:
            return None
        if values[0] == "..":
            return None
        return parse_datetime(values[0])

    @property
    def end_date(self) -> Optional[datetime]:
        values = self.datetime.split("/")
        if len(values) == 1:
            return parse_datetime(values[0])
        if values[1] == "..":
            return None
        return parse_datetime(values[1])

    @validator("intersects")
    def validate_spatial(cls, v, values):
        if v and values["bbox"]:
            raise ValueError("intersects and bbox parameters are mutually exclusive")
        return v

    @validator("bbox")
    def validate_bbox(cls, v: BBox):
        if v:
            # Validate order
            if len(v) == 4:
                xmin, ymin, xmax, ymax = v
            else:
                xmin, ymin, min_elev, xmax, ymax, max_elev = v
                if max_elev < min_elev:
                    raise ValueError(
                        "Maximum elevation must greater than minimum elevation"
                    )

            if xmax < xmin:
                raise ValueError(
                    "Maximum longitude must be greater than minimum longitude"
                )

            if ymax < ymin:
                raise ValueError(
                    "Maximum longitude must be greater than minimum longitude"
                )

            # Validate against WGS84
            if xmin < -180 or ymin < -90 or xmax > 180 or ymax > 90:
                raise ValueError("Bounding box must be within (-180, -90, 180, 90)")

        return v

    @validator("datetime")
    def validate_datetime(cls, v):
        if "/" in v:
            values = v.split("/")
        else:
            # Single date is interpreted as end date
            values = ["..", v]

        dates = []
        for value in values:
            if value == "..":
                dates.append(value)
                continue

            parse_datetime(value)
            dates.append(value)

        if ".." not in dates:
            if parse_datetime(dates[0]) > parse_datetime(dates[1]):
                raise ValueError(
                    "Invalid datetime range, must match format (begin_date, end_date)"
                )

        return v

    @property
    def spatial_filter(self) -> Optional[_GeometryBase]:
        """Return a geojson-pydantic object representing the spatial filter for the search request.

        Check for both because the ``bbox`` and ``intersects`` parameters are mutually exclusive.
        """
        if self.bbox:
            return Polygon(
                coordinates=[
                    [
                        [self.bbox[0], self.bbox[3]],
                        [self.bbox[2], self.bbox[3]],
                        [self.bbox[2], self.bbox[1]],
                        [self.bbox[0], self.bbox[1]],
                        [self.bbox[0], self.bbox[3]],
                    ]
                ]
            )
        if self.intersects:
            return self.intersects
        return



from enum import auto
from types import DynamicClassAttribute
from typing import Any, Callable

from stac_pydantic.utils import AutoValueEnum

# TODO: These are defined in the spec but aren't currently implemented by the operator syntax
UNSUPPORTED_OPERATORS = {"startsWith", "endsWith", "contains", "in"}

_OPERATIONS = {
    "eq": lambda x, y: x == y,
    "ne": lambda x, y: x != y,
    "lt": lambda x, y: x < y,
    "le": lambda x, y: x <= y,
    "gt": lambda x, y: x > y,
    "ge": lambda x, y: x >= y,
    "startsWith": lambda x, y: x.startsWith(y),
    "endsWith": lambda x, y: x.endsWith(y),
    "contains": lambda x, y: y in x,
}


class Operator(str, AutoValueEnum):
    """
    https://github.com/radiantearth/stac-api-spec/tree/master/extensions/query#query-api-extension
    """

    eq = auto()
    ne = auto()
    lt = auto()
    le = auto()
    gt = auto()
    ge = auto()
    startsWith = auto()
    endsWith = auto()
    contains = auto()

    @DynamicClassAttribute
    def operator(self) -> Callable[[Any, Any], bool]:
        """Return python operator"""
        return _OPERATIONS[self._value_]



from typing import List, Optional

from pydantic import AnyUrl, BaseModel, Field, constr, root_validator

from stac_pydantic.links import Links
from stac_pydantic.version import STAC_VERSION


class Catalog(BaseModel):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/catalog-spec/catalog-spec.md
    """

    id: constr(min_length=1)
    description: constr(min_length=1)
    stac_version: constr(min_length=1) = Field(STAC_VERSION, const=True)
    links: Links
    stac_extensions: Optional[List[AnyUrl]]
    title: Optional[str]
    type: constr(min_length=1) = Field("Catalog", const=True)

    class Config:
        use_enum_values = True
        extra = "allow"

    def to_dict(self, **kwargs):
        return self.dict(by_alias=True, exclude_unset=True, **kwargs)

    def to_json(self, **kwargs):
        return self.json(by_alias=True, exclude_unset=True, **kwargs)



from datetime import datetime
from enum import Enum, auto
from typing import List, Optional, Tuple, Union

from pydantic import BaseModel, Extra, Field, confloat, constr

from stac_pydantic.utils import AutoValueEnum

NumType = Union[float, int]
BBox = Union[
    Tuple[NumType, NumType, NumType, NumType],  # 2D bbox
    Tuple[NumType, NumType, NumType, NumType, NumType, NumType],  # 3D bbox
]

# https://tools.ietf.org/html/rfc3339#section-5.6
# Unused, but leaving it here since it's used by dependencies
DATETIME_RFC339 = "%Y-%m-%dT%H:%M:%SZ"


class MimeTypes(str, Enum):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/item-spec/item-spec.md#media-types
    """

    # Raster
    geotiff = "image/tiff; application=geotiff"
    cog = "image/tiff; application=geotiff; profile=cloud-optimized"
    jp2 = "image/jp2"
    png = "image/png"
    jpeg = "image/jpeg"
    # Vector
    geojson = "application/geo+json"
    geopackage = "application/geopackage+sqlite3"
    kml = "application/vnd.google-earth.kml+xml"
    kmz = "application/vnd.google-earth.kmz"
    # Others
    hdf = "application/x-hdf"
    hdf5 = "application/x-hdf5"
    xml = "application/xml"
    json = "application/json"
    html = "text/html"
    text = "text/plain"


class AssetRoles(str, AutoValueEnum):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/extensions/asset/README.md
    """

    thumbnail = auto()
    overview = auto()
    data = auto()
    metadata = auto()


class ProviderRoles(str, AutoValueEnum):
    licensor = auto()
    producer = auto()
    processor = auto()
    host = auto()


class Provider(BaseModel):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/collection-spec/collection-spec.md#provider-object
    """

    name: constr(min_length=1)
    description: Optional[str]
    roles: Optional[List[str]]
    url: Optional[str]


class StacCommonMetadata(BaseModel):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/item-spec/common-metadata.md#date-and-time-range
    """

    title: Optional[str] = Field(None, alias="title")
    description: Optional[str] = Field(None, alias="description")
    start_datetime: Optional[datetime] = Field(None, alias="start_datetime")
    end_datetime: Optional[datetime] = Field(None, alias="end_datetime")
    created: Optional[datetime] = Field(None, alias="created")
    updated: Optional[datetime] = Field(None, alias="updated")
    platform: Optional[str] = Field(None, alias="platform")
    instruments: Optional[List[str]] = Field(None, alias="instruments")
    constellation: Optional[str] = Field(None, alias="constellation")
    mission: Optional[str] = Field(None, alias="mission")
    providers: Optional[List[Provider]] = Field(None, alias="providers")
    gsd: Optional[confloat(gt=0)] = Field(None, alias="gsd")

    class Config:
        json_encoders = {datetime: lambda v: v.strftime(DATETIME_RFC339)}


class Asset(StacCommonMetadata):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/item-spec/item-spec.md#asset-object
    """

    href: constr(min_length=1)
    type: Optional[str]
    title: Optional[str]
    description: Optional[str]
    roles: Optional[List[str]]

    class Config:
        allow_population_by_field_name = True
        use_enum_values = True
        extra = Extra.allow



from typing import Dict, Optional, Set

from pydantic import BaseModel


class FieldsExtension(BaseModel):
    """
    https://github.com/radiantearth/stac-api-spec/tree/master/extensions/fields#fields-api-extension
    """

    includes: Optional[Set[str]]
    excludes: Optional[Set[str]]

    def _get_field_dict(self, fields: Set[str]) -> Dict:
        """Internal method to create a dictionary for advanced include or exclude of pydantic fields on model export

        Ref: https://pydantic-docs.helpmanual.io/usage/exporting_models/#advanced-include-and-exclude
        """
        field_dict = {}
        for field in fields:
            if "." in field:
                parent, key = field.split(".")
                if parent not in field_dict:
                    field_dict[parent] = {key}
                else:
                    field_dict[parent].add(key)
            else:
                field_dict[field] = ...  # type:ignore
        return field_dict

    @property
    def filter(self) -> Dict:
        """
        Create dictionary of fields to include/exclude on model export based on the included and excluded fields passed
        to the API.  The output of this property may be passed to pydantic's serialization methods to include or exclude
        certain keys.

        Ref: https://pydantic-docs.helpmanual.io/usage/exporting_models/#advanced-include-and-exclude
        """
        include = set()
        # If only include is specified, add fields to the set
        if self.includes and not self.excludes:
            include = include.union(self.includes)
        # If both include + exclude specified, find the difference between sets
        elif self.includes and self.excludes:
            include = include.union(self.includes) - self.excludes
        return {
            "include": self._get_field_dict(include),
            "exclude": self._get_field_dict(self.excludes),
        }



from dataclasses import dataclass
from typing import ClassVar, Tuple
from urllib.parse import urljoin

from stac_pydantic.links import Link, Links, Relations
from stac_pydantic.shared import MimeTypes


@dataclass
class BaseLinks:
    """Create inferred links common to collections and items."""

    base_url: str
    _link_members: ClassVar[Tuple[str]] = ("root",)

    def root(self) -> Link:
        """Return the catalog root."""
        return Link(
            rel=Relations.root, type=MimeTypes.json, href=urljoin(self.base_url, "/")
        )

    def create_links(self) -> Links:
        """Create inferred links"""
        return Links.parse_obj(
            [getattr(self, member)() for member in self._link_members]
        )


@dataclass
class CollectionLinks(BaseLinks):
    """Create inferred links specific to collections."""

    collection_id: str
    _link_members: ClassVar[Tuple[str]] = ("root", "self", "parent", "item")

    def self(self) -> Link:
        """Create the `self` link."""
        return Link(
            rel=Relations.self,
            type=MimeTypes.json,
            href=urljoin(self.base_url, f"/collections/{self.collection_id}"),
        )

    def parent(self) -> Link:
        """Create the `parent` link."""
        return Link(
            rel=Relations.parent, type=MimeTypes.json, href=urljoin(self.base_url, "/")
        )

    def item(self) -> Link:
        """Create the `item` link."""
        return Link(
            rel=Relations.item,
            type=MimeTypes.geojson,
            href=urljoin(self.base_url, f"/collections/{self.collection_id}/items"),
        )


@dataclass
class ItemLinks(BaseLinks):
    """Create inferred links specific to items."""

    collection_id: str
    item_id: str
    _link_members: ClassVar[Tuple[str]] = ("root", "self", "parent", "collection")

    def self(self) -> Link:
        """Create the `self` link."""
        return Link(
            rel=Relations.self,
            type=MimeTypes.geojson,
            href=urljoin(
                self.base_url, f"/collections/{self.collection_id}/items/{self.item_id}"
            ),
        )

    def parent(self) -> Link:
        """Create the `parent` link."""
        return Link(
            rel=Relations.parent,
            type=MimeTypes.json,
            href=urljoin(self.base_url, f"/collections/{self.collection_id}"),
        )

    def collection(self) -> Link:
        """Create the `collection` link."""
        return Link(
            rel=Relations.collection,
            type=MimeTypes.json,
            href=urljoin(self.base_url, f"/collections/{self.collection_id}"),
        )



from datetime import datetime as dt
from typing import Dict, List, Optional, Union

from geojson_pydantic.features import Feature, FeatureCollection
from pydantic import AnyUrl, Field, constr, validator
from pydantic.datetime_parse import parse_datetime

from stac_pydantic.api.extensions.context import ContextExtension
from stac_pydantic.links import Links
from stac_pydantic.shared import DATETIME_RFC339, Asset, BBox, StacCommonMetadata
from stac_pydantic.version import STAC_VERSION


class ItemProperties(StacCommonMetadata):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/item-spec/item-spec.md#properties-object
    """

    datetime: Union[dt, str] = Field(..., alias="datetime")

    @validator("datetime")
    def validate_datetime(cls, v, values):
        if v == "null":
            if not values["start_datetime"] and not values["end_datetime"]:
                raise ValueError(
                    "start_datetime and end_datetime must be specified when datetime is null"
                )

        if isinstance(v, str):
            return parse_datetime(v)

        return v

    class Config:
        extra = "allow"
        json_encoders = {dt: lambda v: v.strftime(DATETIME_RFC339)}


class Item(Feature):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/item-spec/item-spec.md
    """

    id: constr(min_length=1)
    stac_version: constr(min_length=1) = Field(STAC_VERSION, const=True)
    properties: ItemProperties
    assets: Dict[str, Asset]
    links: Links
    bbox: BBox
    stac_extensions: Optional[List[AnyUrl]]
    collection: Optional[str]

    def to_dict(self, **kwargs):
        return self.dict(by_alias=True, exclude_unset=True, **kwargs)

    def to_json(self, **kwargs):
        return self.json(by_alias=True, exclude_unset=True, **kwargs)


class ItemCollection(FeatureCollection):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/item-spec/itemcollection-spec.md
    """

    stac_version: constr(min_length=1) = Field(STAC_VERSION, const=True)
    features: List[Item]
    stac_extensions: Optional[List[AnyUrl]]
    links: Links
    context: Optional[ContextExtension]

    def to_dict(self, **kwargs):
        return self.dict(by_alias=True, exclude_unset=True, **kwargs)

    def to_json(self, **kwargs):
        return self.json(by_alias=True, exclude_unset=True, **kwargs)
